@@ -34,7 +34,7 @@ import nice.tools.util.Chronometer;
 /**
    Static class performing the coverage and non-ambiguity test.
    
-   @version $Date: 2003/10/23 14:49:52 $
+   @version $Date: 2004/02/25 11:23:29 $
    @author Daniel Bonniot
  */
 
@@ -99,7 +99,7 @@ public final class Dispatch
   {
     Stack sortedAlternatives = Alternative.sortedAlternatives(m);
     
-    if (! trivialTestOK(sortedAlternatives))
+    if (! trivialTestJava(m, sortedAlternatives))
       test(m, sortedAlternatives, true);
     
     if(Debug.codeGeneration)
@@ -122,6 +122,28 @@ public final class Dispatch
     return true;
   }
 
+  private static boolean trivialTestJava(JavaMethod m, Stack alternatives)
+  {
+    gnu.bytecode.Method reflectMethod = m.reflectMethod;
+
+    // Static methods and constructors cannot be overriden, so there is
+    // no problem.
+    if (reflectMethod.getStaticFlag() || reflectMethod.isConstructor())
+      return true;
+
+    if (! reflectMethod.isAbstract())
+      // The only risk with a non abstract method is that it might be
+      // ambiguous.
+      // We know this won't happen in two cases:
+      //   1) if there is only one argument (single
+      //      inheritance, since the root must be a class), 
+      //   2) if there are less than two implementations.
+      return m.getArity() == 1 || alternatives.size() < 2;
+
+    // This method will need testing.
+    return false;
+  }
+
   private static boolean[] findUsedPositions(int len, Stack alternatives)
   {
     // We want all solutions for both the nullness marker and the tag,
@@ -166,7 +188,7 @@ public final class Dispatch
     boolean[] isValue = new boolean[method.getArity()];
     List values = generateValues(sortedAlternatives, isValue);
     boolean hasValues = values.size() > 0;
-   
+    List errors = new ArrayList(3); 
     int nb_errors = 0;
     for(Iterator i = multitags.iterator(); i.hasNext();)
       {
@@ -182,20 +204,21 @@ public final class Dispatch
 	      continue;
 	  }
 	
-	if (test(method, tags, sortedAlternatives, firstArg))
+	if (test(method, tags, sortedAlternatives, firstArg, errors))
 	{
 	  if (++nb_errors > 3)
 	    break;
         }
 	else if (hasValues && 
-	      testValues(method, tags, values, isValue, sortedAlternatives) )
+	      testValues(method, tags, values, isValue, sortedAlternatives, errors) )
 	  if (++nb_errors > 0)
 	    break;
 
       }
     if (nb_errors > 0)
       User.error(method, "The implementation test failed for method " + 
-		 method.getName());
+		 method.toString() + ":\n" + 
+		 Util.map("", "\n", "", errors));
   }
 
   private static ClassType classTypeOfNiceClass(TypeConstructor tc)
@@ -220,7 +243,8 @@ public final class Dispatch
   private static boolean test(MethodDeclaration method, 
 			      TypeConstructor[] tags, 
 			      final Stack sortedAlternatives,
-			      ClassType firstArg)
+			      ClassType firstArg,
+			      List errors)
   {
     boolean failed = false;
 
@@ -242,10 +266,7 @@ public final class Dispatch
 		   !a.containsTypeMatchingValue())
 	    {
 	      failed = true;
-	      User.warning
-		(method,
-		 "Ambiguity for method "+method+
-		 "\nFor parameters of type " + toString(tags)+
+	      errors.add("ambiguity for parameters of type " + toString(tags)+
 		 "\nboth\n" + first.printLocated() + 
 		 "\nand\n" + a.printLocated() + "\nmatch.");
 	    }
@@ -268,10 +289,7 @@ public final class Dispatch
 	    (method, "Method " + method + " is declared but never implemented:\n" +
 	     "no alternative matches " + toString(tags));
 	else
-	  User.warning(method,
-		       "Method " + method + " is not completely covered:\n" + 
-		       "no alternative matches " + 
-		       toString(tags));
+	  errors.add("no alternative matches " + toString(tags));
       }
     return failed;
   }
@@ -284,7 +302,8 @@ public final class Dispatch
      TypeConstructor[] tags,
      List valueCombis,
      boolean[] isValue, 
-     final Stack sortedAlternatives)
+     final Stack sortedAlternatives,
+     List errors)
   {
     boolean failed = false;
     List sortedTypeMatches = new ArrayList();
@@ -309,10 +328,7 @@ public final class Dispatch
 	  else if (!Alternative.less(first, a))
 	    {
 	      failed = true;
-	      User.warning
-		(method,
-		 "Ambiguity for method "+method+
-		 "\nFor parameters of type/value " + toString(tags, values, isValue)+
+	      errors.add("ambiguity for parameters of type/value " + toString(tags, values, isValue)+
 		 "\nboth\n" + first.printLocated() + 
 		 "\nand\n" + a.printLocated() + "\nmatch.");
               break outer;
@@ -321,10 +337,7 @@ public final class Dispatch
       if(first==null)
       {
 	failed = true;
-	User.warning(method,
-		       "Method " + method + " is not completely covered:\n" + 
-		       "no alternative matches " + 
-		       toString(tags, values, isValue));
+	errors.add("no alternative matches "+ toString(tags, values, isValue));
 	break;
       }
     }
@@ -392,7 +405,7 @@ public final class Dispatch
 	  {
 	    MonotypeConstructor mc = (MonotypeConstructor) arg;
 	    marker = mc.getTC();
-	    raw = nice.tools.code.Types.rawType(mc);
+	    raw = nice.tools.typing.Types.rawType(mc);
 	  }
 	else
 	  {
