@@ -16,6 +16,7 @@ import bossa.util.*;
 import bossa.syntax.*;
 import gnu.bytecode.*;
 import gnu.expr.*;
+import nice.tools.code.Import;
 
 import java.util.*;
 import java.util.jar.*;
@@ -27,7 +28,7 @@ import gnu.expr.ClassExp;
 /**
    A Nice package.
    
-   @version $Date: 2004/02/16 21:47:12 $
+   @version $Date: 2004/06/11 15:58:58 $
    @author Daniel Bonniot (bonniot@users.sourceforge.net)
  */
 public class Package implements mlsub.compilation.Module, Located, bossa.syntax.Module
@@ -328,15 +329,15 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
   {
     if (!isRoot)
       return;
-    
+
     // If at least one package is recompiled, the root will also be recompiled
     if (compiling())
       {
         compilation.progress(this, "linking");
 	bossa.link.Dispatch.test(this);
-    
+
 	finishCompilation();
-    
+
 	compilation.exitIfErrors();
       }
 
@@ -426,10 +427,11 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
 	  f.print("import "+m.getName()+";\n");
 	}
       f.println();
-    
+
       for(int i = 0; i < opens.length; i++)
 	{
-	  f.print("import " + opens[i] + ".*;\n");
+	  f.print("import " + opens[i] + ".*" + 
+		   (Import.isStrictPackage(opens[i]) ? "(!)" : "") + ";\n");
 	}
       f.println();
 
@@ -636,44 +638,31 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
       classe.outer = getImplementationClass();
   }
 
-  public gnu.expr.Declaration addGlobalVar(String name, Type type, boolean constant)
+  public void addGlobalVar(gnu.expr.Declaration decl, 
+                           boolean constant)
   {
-    gnu.expr.Declaration declaration = new gnu.expr.Declaration(name, type);
-
     if (!compiling())
       // The code is already there
       {
-	declaration.setSimple(false);
-	declaration.field = source.getBytecode().getField(name);
+	decl.setSimple(false);
+	decl.field = source.getBytecode().getField(decl.getName());
 
-	if (declaration.field == null)
+	if (decl.field == null)
 	  Internal.error(this,
-			 "The compiled file is not consistant with the interface file for global variable " + name);
+			 "The compiled file is not consistant with the interface file for global variable " + decl.getName());
       }
     else
       {
-	getImplementationClass().addDeclaration(declaration);
-        if (constant) declaration.setFlag(Declaration.IS_CONSTANT);
+	getImplementationClass().addDeclaration(decl);
+        if (constant) decl.setFlag(Declaration.IS_CONSTANT);
 
-	declaration.setFlag(Declaration.STATIC_SPECIFIED|Declaration.TYPE_SPECIFIED);
+	decl.setFlag(Declaration.STATIC_SPECIFIED|Declaration.TYPE_SPECIFIED);
       }
-    
-    return declaration;
   }
 
   private gnu.expr.Package thisPkg;
   private ClassExp implementationClass, dispatchClass;
 
-  private static ModuleExp createModule(String name)
-  {
-    ModuleExp res = new ModuleExp();
-    res.setName(name);
-    res.body = QuoteExp.voidExp;
-    res.setFlag(ModuleExp.STATIC_SPECIFIED);
-    res.setSuperType(gnu.bytecode.Type.pointer_type);
-    return res;
-  }
-
   private ClassExp createClassExp(String name)
   {
     ClassExp res = new ClassExp();
@@ -724,7 +713,7 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
   private void generateCode()
   {
     if (compiling())
-      compilation.progress(this, "generating code");    
+      compilation.progress(this, "generating code");
     
     ast.compile(compiling());
   }
@@ -736,7 +725,7 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
     // However if the class exists but is invalid, we create a new one.
 
     String className = this.name + "." + name;
-    ClassType res; 
+    ClassType res;
     try{
       res = ClassType.make(className);
     }
@@ -755,7 +744,7 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
   private Method lookupClassMethod(ClassType clas, String name,
 				   String attribute, String value)
   {
-    if (clas == null) 
+    if (clas == null)
       return null;
 
     name = nice.tools.code.Strings.escape(name);
@@ -766,7 +755,7 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
 	  // in the bytecode by appending "$..."
           // but names appended with "$$..." may not be matched because
           // that are escape characters
-	  || m.getName().startsWith(name) 
+	  || m.getName().startsWith(name)
 	  && m.getName().charAt(name.length()) == '$'
           && m.getName().charAt(name.length()+1) != '$')
 	{
@@ -794,13 +783,13 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
         This would not be the case if we recomputed it,
 	as the precise types are found during typechecking.
     */
-    Method meth = lookupClassMethod(source.getDispatch(), name, 
+    Method meth = lookupClassMethod(source.getDispatch(), name,
 				    "id", def.getFullName());
     if (meth != null) // Reuse existing dispatch method header
       {
 	// The dispatch code will have to be regenerated anyway
 	meth.eraseCode();
-	
+
 	argTypes = meth.arg_types;
 	retType  = meth.return_type;
         // Make sure we use the same bytecode name, since compiled code
@@ -812,15 +801,37 @@ public class Package implements mlsub.compilation.Module, Located, bossa.syntax.
 	argTypes = def.javaArgTypes();
 	retType  = def.javaReturnType();
       }
+
+    // Try to compile the dispatch method as a member method if possible.
+    NiceClass receiver;
+    if (argTypes.length == 0)
+      receiver = null;
+    else
+      {
+        receiver = NiceClass.get(def.getArgTypes()[0]);
+
+        // JVM interfaces cannot contain code.
+        if (receiver != null && receiver.isInterface())
+          receiver = null;
+
+        // Special treatment for serialization at the moment.
+        if (def.getArity() == 2 && 
+            (name.equals("writeObject")||name.equals("readObject")))
+          receiver = null;
+      }
+
     res = nice.tools.code.Gen.createMethod
-      (name, argTypes, retType, def.getSymbols());
+      (name, argTypes, retType, def.getSymbols(), true, receiver != null);
     res.parameterCopies = def.formalParameters().getParameterCopies();
 
     // add unique information to disambiguate which method this represents
     res.addBytecodeAttribute
       (new MiscAttr("id", def.getFullName().getBytes()));
 
-    return addMethod(res, false);
+    if (receiver != null)
+      return receiver.addJavaMethod(res);
+    else
+      return addMethod(res, false);
   }
 
   /** Add a method to this package and return an expression to refer it. */
