@@ -29,7 +29,7 @@ import gnu.expr.QuoteExp;
 /**
    Abstract syntax for a class definition.
    
-   @version $Date: 2003/12/22 01:30:59 $
+   @version $Date: 2004/06/11 15:53:35 $
    @author Daniel Bonniot
  */
 public class NiceClass extends ClassDefinition.ClassImplementation
@@ -81,6 +81,11 @@ public class NiceClass extends ClassDefinition.ClassImplementation
 
   ClassDefinition definition;
 
+  public boolean isInterface()
+  {
+    return definition instanceof ClassDefinition.Interface;
+  }
+
   private static NewField[] noFields = new NewField[0];
   private static OverridenField[] noOverrides = new OverridenField[0];
   
@@ -93,6 +98,14 @@ public class NiceClass extends ClassDefinition.ClassImplementation
       return null;
   }
 
+  public static NiceClass get(mlsub.typing.Monotype type)
+  {
+    if (! nice.tools.typing.Types.isSure(type))
+      return null;
+
+    return get(nice.tools.typing.Types.constructor(type));
+  }
+
   NiceClass getParent()
   {
     return get(definition.getSuperClass());
@@ -119,7 +132,7 @@ public class NiceClass extends ClassDefinition.ClassImplementation
     (MonoSymbol sym, Expression value, 
      boolean isFinal, boolean isTransient, boolean isVolatile, String docString)
   {
-    if (definition instanceof ClassDefinition.Interface)
+    if (isInterface())
       User.error(sym, "An interface cannot have a field.");
 
     return new NewField(sym, value, isFinal, isTransient, isVolatile, docString);
@@ -127,7 +140,7 @@ public class NiceClass extends ClassDefinition.ClassImplementation
 
   public Field makeOverride (MonoSymbol sym, Expression value)
   {
-    if (definition instanceof ClassDefinition.Interface)
+    if (isInterface())
       User.error(sym, "An interface cannot have a field.");
 
     return new OverridenField(sym, value);
@@ -158,7 +171,7 @@ public class NiceClass extends ClassDefinition.ClassImplementation
     {
       sym.type = sym.syntacticType.resolve(typeScope);
       
-      if (Types.isVoid(sym.type))
+      if (nice.tools.typing.Types.isVoid(sym.type))
 	User.error(sym, "A field cannot have void type");
 
       value = dispatch.analyse(value, scope, typeScope);
@@ -660,7 +673,7 @@ public class NiceClass extends ClassDefinition.ClassImplementation
      on the type parameters.
   */
   private static List getNativeConstructorParameters
-    (TypeConstructor tc, List constraints, TypeSymbol[] binders)
+    (TypeConstructor tc, List constraints)
   {
     List constructors = TypeConstructors.getConstructors(tc);
     if (constructors == null)
@@ -689,21 +702,24 @@ public class NiceClass extends ClassDefinition.ClassImplementation
   }
 
   private List getParentConstructorParameters
-    (List constraints, TypeSymbol[] binders)
+    (List constraints, mlsub.typing.Monotype[] typeParameters)
   {
     TypeScope scope = Node.getGlobalTypeScope();
     Map map = null;
-    if (binders != null)
+    if (typeParameters != null)
       {
         // Constructs a type scope that maps the type parameters of this
         // class to the corresponding symbol in the constructor.
 	scope = new TypeScope(scope);
         map = new HashMap();
-        TypeSymbol[] ourBinders = definition.getBinders();
-	for (int i = 0; i < binders.length; i++)
+        mlsub.typing.Monotype[] ourTypeParameters = 
+          definition.getTypeParameters();
+	for (int i = 0; i < ourTypeParameters.length; i++)
 	  try {
-	    scope.addMapping(ourBinders[i].toString(), binders[i]);
-	    map.put(ourBinders[i], binders[i]);
+            TypeSymbol ourSym = asTypeSymbol(ourTypeParameters[i]);
+            TypeSymbol sym = asTypeSymbol(typeParameters[i]);
+            scope.addMapping(ourSym.toString(), sym);
+            map.put(ourSym, sym);
 	  } catch(TypeScope.DuplicateName e) {}
       }
 
@@ -733,16 +749,25 @@ public class NiceClass extends ClassDefinition.ClassImplementation
     return res;
   }
 
-  private List getConstructorParameters(List constraints, TypeSymbol[] binders)
+  private TypeSymbol asTypeSymbol(mlsub.typing.Monotype type)
+  {
+    if (type instanceof TypeSymbol)
+      return (TypeSymbol) type;
+    else
+      return ((MonotypeConstructor) type).getTC();
+  }
+
+  private List getConstructorParameters(List constraints, 
+                                        mlsub.typing.Monotype[] typeParameters)
   {
     TypeConstructor supTC = definition.getSuperClass();
 
     NiceClass sup = get(supTC);
     List res;
     if (sup == null)
-      res = getNativeConstructorParameters(supTC, constraints, binders);
+      res = getNativeConstructorParameters(supTC, constraints);
     else
-      res = sup.getParentConstructorParameters(constraints, binders);
+      res = sup.getParentConstructorParameters(constraints, typeParameters);
 
     if (overrides.length > 0)
       for (Iterator i = res.iterator(); i.hasNext();)
@@ -796,7 +821,7 @@ public class NiceClass extends ClassDefinition.ClassImplementation
       // The constructors are loaded from the compiled package.
       return;
 
-    if (definition instanceof ClassDefinition.Interface)
+    if (isInterface())
       return;
 
     List constraints;
@@ -806,7 +831,10 @@ public class NiceClass extends ClassDefinition.ClassImplementation
     else 
       constraints = new LinkedList();
 
-    List allConstructorParams = getConstructorParameters(constraints, binders);
+    mlsub.typing.Monotype[] typeParameters = definition.getTypeParameters();
+
+    List allConstructorParams = 
+      getConstructorParameters(constraints, typeParameters);
 
     Constraint cst;
     if (binders != null)
@@ -865,6 +893,11 @@ public class NiceClass extends ClassDefinition.ClassImplementation
     if (constructorMethod != null)
       for (int i = 0; i < constructorMethod.length; i++)
 	constructorMethod[i].getCode();
+
+    // Take into account external interface implementations, which 
+    // can add new interfaces to implement in the bytecode.
+    classe.supers = computeSupers();
+    classe.recomputeInterfaces();
   }
 
   private gnu.expr.Expression typeExpression(TypeConstructor tc)
