@@ -23,12 +23,12 @@ import mlsub.typing.FunType;
 import mlsub.typing.TupleType;
 import mlsub.typing.Monotype;
 
-import nice.tools.code.Types;
+import nice.tools.typing.Types;
 
 /**
    A symbol, for which overloading resolution has yet to be done.
 
-   @version $Date: 2004/02/03 10:20:48 $
+   @version $Date: 2004/06/23 21:27:46 $
    @author Daniel Bonniot (d.bonniot@mail.dotcom.fr)
 */
 public class OverloadedSymbolExp extends Expression
@@ -150,15 +150,17 @@ public class OverloadedSymbolExp extends Expression
 	if(Debug.overloading) 
 	  Debug.println("Overloading: Trying with "+s);
 
-	Polytype[] argsType = 
-	  Expression.getType(arguments.getExpressions(s));
-
 	// we clone the type to avoid clashes with another use
 	// of the same symbol
 	// the cloned type is stored in the VarSymbol
 	// and we check that cloneType() is not called twice
 	// before the clone type is released
-	s.makeClonedType(argsType, arguments.getUsedArguments(s));
+	s.makeClonedType();
+
+        Polytype[] argsType = 
+          computeArgsType(arguments.getExpressions(s),
+                          s.getClonedType(), arguments.getUsedArguments(s));
+
 	Polytype t = CallExp.wellTyped(s.getClonedType(), argsType);
 
 	if (t == null)
@@ -195,15 +197,16 @@ public class OverloadedSymbolExp extends Expression
       }
 
     removeNonMinimal(symbols, arguments);
+    removeOverlappingJavaMethods(symbols);
     
     if (symbols.size() == 1)
       {
 	VarSymbol res = (VarSymbol) symbols.get(0);
 	// store the formal argument types for later use together with the type
-	callExp.argTypes = nice.tools.code.Types.domain(res.getClonedType());
+        callExp.setComputedType((Polytype) arguments.types.get(res),
+                                nice.tools.typing.Types.parameters(res.getClonedType()));
 	res.releaseClonedType();
 
-	callExp.type = (Polytype) arguments.types.get(res);
 	// store the expression (including default arguments)
 	callExp.arguments.computedExpressions = arguments.getExpressions(res);
 	//callExp.arguments = null; // free memory
@@ -214,6 +217,40 @@ public class OverloadedSymbolExp extends Expression
     throw new AmbiguityError();
   }
 
+  private Polytype[] computeArgsType(Expression[] args,
+                                     Polytype functionType,
+                                     int[] usedArguments)
+  {
+    Polytype[] res = new Polytype[args.length];
+
+    mlsub.typing.Monotype[] domain = null;
+
+    /* Where a default value was used, use the declared argument type instead
+       of the value's type. This is more robust, as the application type will
+       not depend on the default value.
+       Furthermore, this avoids running into problems when the default value
+       refers to type parameters (in anonymous functions, by refering to
+       previous arguments, ...) which would not be in sync with the cloned
+       ones.
+       This is only needed when the type is polymorphic.
+    */
+    for (int i = 0; i < res.length; i++)
+      if (usedArguments != null && usedArguments[i] == 0)
+        {
+          if (domain == null)
+            {
+              mlsub.typing.Monotype fun = Types.rawType(functionType.getMonotype());
+              domain = ((mlsub.typing.FunType) fun).domain();
+            }
+
+          res[i] = new Polytype(domain[i]);
+        }
+      else
+        res[i] = args[i].getType();
+
+    return res;
+  }
+
   Expression resolveOverloading(Polytype expectedType)
   {
     if(Debug.overloading) 
@@ -230,7 +267,7 @@ public class OverloadedSymbolExp extends Expression
     for (Iterator i = symbols.iterator(); i.hasNext();)
       {
 	VarSymbol s = (VarSymbol) i.next();
-	s.makeClonedType(null, null);
+	s.makeClonedType();
 	try{
 	  Typing.leq(s.getClonedType(), expectedType);
 	  if(Debug.overloading)
@@ -260,7 +297,7 @@ public class OverloadedSymbolExp extends Expression
       if (res != null)
         return res;
 
-      if (Types.domain(expectedType) != null)
+      if (Types.parameters(expectedType) != null)
         { // in case of function objects symbols find the most precise match
 	  List nonMin = removeNonMinimal(symbols);
 	  if (symbols.size() == 1)
@@ -382,7 +419,7 @@ public class OverloadedSymbolExp extends Expression
     
     for(int s1 = 0; s1<len; s1++) {
 
-      Domain d1 = domain(syms[s1].getType());
+      Domain d1 = Types.domain(syms[s1].getType());
       
       for(int s2 = 0; s2<len; s2++)
 	/*
@@ -397,7 +434,7 @@ public class OverloadedSymbolExp extends Expression
 	*/
 	if (s1 != s2 && !remove[s2]) {
 
-	  Domain d2 = domain(syms[s2].getType());
+	  Domain d2 = Types.domain(syms[s2].getType());
 
 	  try{
 	    Typing.leq(d2, d1);
@@ -498,7 +535,7 @@ public class OverloadedSymbolExp extends Expression
   private static Domain domain(Polytype t, int[] usedArguments)
   {
     // remove nullness marker
-    Monotype[] m = Types.domain(t.getMonotype());
+    Monotype[] m = Types.parameters(t.getMonotype());
 
     Monotype[] dom;
 
@@ -519,15 +556,47 @@ public class OverloadedSymbolExp extends Expression
 
       }
 
-    return new Domain(t.getConstraint(), new TupleType(dom));
+    return new Domain(t.getConstraint(), dom);
   }
 
-  private static Domain domain(Polytype t)
+  /**
+     When a Java class implements two interfaces,
+     there is no ambiguity between methods of the same signature
+     defined in both interfaces, since they are implemented by the same
+     method in the class. So this method removes all but one of the 
+     overlapping methods.
+  */
+  private static void removeOverlappingJavaMethods(List symbols)
   {
-    // remove nullness marker
-    Monotype[] m = Types.domain(t.getMonotype());
+    // optimization
+    if(symbols.size()<2)
+      return;
+    
+    int len = symbols.size();
+    VarSymbol[] syms = (VarSymbol[])
+      symbols.toArray(new VarSymbol[len]);
 
-    return new Domain(t.getConstraint(), new TupleType(m));
+    for (int i = 0; i < syms.length; i++)
+      for (int j = i+1; j < syms.length; j++)
+        if (overlappingJavaMethods(syms[i].getMethodDeclaration(),
+                                   syms[j].getMethodDeclaration()))
+          {
+            // We can remove either, since they lead to the same implementation
+            symbols.remove(syms[i]);
+            // Since we removed i, we don't need to continue this loop.
+            // This would not be true if we decided to remove j.
+            break;
+          }
+  }
+
+  private static boolean overlappingJavaMethods(MethodDeclaration m1,
+                                                MethodDeclaration m2)
+  {
+    if (! (m1 instanceof JavaMethod) ||
+        ! (m2 instanceof JavaMethod))
+      return false;
+
+    return Arrays.equals(m1.javaArgTypes(), m2.javaArgTypes());
   }
 
   void computeType()
