@@ -18,19 +18,21 @@ import mlsub.typing.*;
 import mlsub.typing.TypeConstructor;
 import mlsub.typing.Monotype;
 import mlsub.typing.MonotypeConstructor;
+import mlsub.typing.Polytype;
 import mlsub.typing.Typing;
 import mlsub.typing.TypingEx;
+import nice.tools.typing.Types;
 
 import gnu.expr.*;
-import nice.tools.code.*;
+import nice.tools.code.Gen;
 
 /**
    Represents the information about one argument of a method body.
    
    @see MethodBodyDefinition
 
-   @version $Date: 2004/02/02 13:08:33 $
-   @author Daniel Bonniot (d.bonniot@mail.dotcom.fr)
+   @version $Date: 2004/06/16 10:00:37 $
+   @author Daniel Bonniot (bonniot@users.sf.net)
 */
 public class Pattern implements Located
 {
@@ -110,6 +112,11 @@ public class Pattern implements Located
     return res;
   }
 
+  public final TypeConstructor getTC()
+  {
+    return tc;
+  }
+
   final TypeConstructor getRuntimeTC()
   {
     return runtimeTC;
@@ -153,8 +160,12 @@ public class Pattern implements Located
     if (def == null)
       return;
 
-    constraint = def.getResolvedConstraint();
-    patternType = new MonotypeConstructor(tc, def.getTypeParameters());
+    Polytype classType = def.getConstrainedType();
+    if (classType != null)
+      {
+        constraint = classType.getConstraint();
+        patternType = classType.getMonotype();
+      }
   }
 
   private static VarSymbol findRefSymbol(LocatedString refName)
@@ -389,6 +400,15 @@ public class Pattern implements Located
     if (that.atNull() || this.atNull())
       return false;
 
+    if (that.atNonNull())
+      // The only reason it could be false is if we are atNull, but that's
+      // already handled.
+      return true;
+
+    // that is not atNonNull nor atAny at this point
+    if (this.atNonNull())
+      return false;
+
     if (that.atBool())
       return this.atBool() && (this.atTrue() == that.atTrue());
 
@@ -430,6 +450,9 @@ public class Pattern implements Located
     if (atNull())
       return tag == PrimitiveType.nullTC;
 
+    if (atNonNull())
+      return tag != PrimitiveType.nullTC;
+
     // a null tc is an unmatchable argument (e.g. function)
     if (tag == null)
       return false;
@@ -560,18 +583,18 @@ public class Pattern implements Located
       return atValue.toString();
 
     if (atAny())
-      return "@_";
+      return name != null ? name.toString() : "Any";
 
     StringBuffer res = new StringBuffer();
-    if (name != null)
-      res.append(name);
-
-    res.append(exactlyAt ? '#' : '@');
+    res.append(exactlyAt ? "#" : (name != null ? "" : "@"));
     
     if (typeConstructor != null)
       res.append(typeConstructor);
     else if (tc != null)
       res.append(tc);
+
+    if (name != null)
+      res.append(" " + name);
     
     return res.toString();
   }
@@ -756,12 +779,15 @@ public class Pattern implements Located
     TypeSymbol sym = Node.getGlobalTypeScope().lookup(name);
 
     if (sym == null)
-      User.error("Pattern " + name + " of method " + methodName + 
-	     " is not known");
+      // This can happen if the class exists only in a later version
+      // of the JDK.
+      throw new Unknown();
 
     return new Pattern((TypeConstructor) sym, exact);
   }
 
+  public static class Unknown extends RuntimeException {}
+
   /****************************************************************
    * Code generation
    ****************************************************************/
@@ -807,7 +833,7 @@ public class Pattern implements Located
     if (atReference())
       return Gen.referenceEqualsExp(atValue.compile(), parameter);
 
-    gnu.bytecode.Type ct = Types.javaType(tc);
+    gnu.bytecode.Type ct = nice.tools.code.Types.javaType(tc);
     if (exactlyAt)
       return Gen.isOfClass(parameter, ct);
       
@@ -819,8 +845,8 @@ public class Pattern implements Located
    ****************************************************************/
   
   LocatedString name,refName;
-  TypeIdent typeConstructor, additional;
-  public TypeConstructor tc;
+  private TypeIdent typeConstructor, additional;
+  TypeConstructor tc;
   TypeConstructor tc2;
   private TypeConstructor runtimeTC;
 
@@ -849,6 +875,12 @@ public class Pattern implements Located
     return atValue != null && !atBool() && !atNull() &&!atIntCompare();
   }
   public boolean atNull() { return atValue instanceof NullExp; }
+  /** This pattern only specifies that the vlaue is not null.
+      This cannot be explicitely used in source programs, but it is useful
+      when a method with a non-null parameter specializes one where that 
+      parameter can be null.
+  */
+  public boolean atNonNull() { return tc == PrimitiveType.sureTC; }
   public boolean atAny()  { return atValue == null && tc == null; }
   public boolean atBool() { 
     return atValue != null && tc == PrimitiveType.boolTC;
