@@ -37,18 +37,25 @@ public abstract class Engine
    * If an enter() completed successfully,
    * a matching leave() MUST be issued some time later by the caller.
    */
-  public static void enter()
+  public static void enter(boolean tentative)
   {
     if(dbg) Debug.println("Enter");
     floating.mark();
     soft.mark();
-    frozenLeqs.mark();
-    for(Iterator i = constraints.iterator();
-	i.hasNext();)
-      { 
-	Engine.Constraint k = (Engine.Constraint)i.next();
-	k.mark();
-	//k.rigidify();
+    formerFree.mark();
+
+    // Once we are in existential mode, we don't mark/backtrack.
+    if (!tentative && existentialLevel > 0)
+      existentialLevel++;
+    else
+      {
+        frozenLeqs.mark();
+
+        for(Iterator i = constraints.iterator(); i.hasNext();)
+          {
+            Engine.Constraint k = (Engine.Constraint)i.next();
+            k.mark();
+          }
       }
   }
   
@@ -95,9 +102,15 @@ public abstract class Engine
    *
    * @exception Unsatisfiable if the constraint was not satisfiable.
    */
-  public static void leave()
+  public static void leave(boolean tentative, boolean commit)
     throws Unsatisfiable
   {
+    // 'tentative' is only meaningful when in existential mode
+    tentative &= existentialLevel > 0;
+    // We only 'commit' in tentative mode
+    commit &= tentative;
+      
+    boolean ok = false;
     try{
       assertFrozens();
 
@@ -118,28 +131,63 @@ public abstract class Engine
 	    throw e;
 	  }
 	}
+      ok = true;
     }
     finally{
-      backtrack();
+      // Even if commit is true, if an error appeared during leaving
+      // then we don't want to keep the changes.
+      backtrack(tentative, ok && commit);
     }
   }
   
-  public static void backtrack()
+  public static void backtrack(boolean tentative, boolean commit)
   {
-    for(Iterator i = constraints.iterator();
-	i.hasNext();)
-      { 
-	Engine.Constraint k = (Engine.Constraint)i.next();
-	k.backtrack();
-      }
     floating.backtrack();
+
+    if (existentialLevel <= 1 || tentative)
+      {
+        for(Iterator i = constraints.iterator();
+            i.hasNext();)
+          { 
+            Engine.Constraint k = (Engine.Constraint) i.next();
+            k.backtrack(commit);
+          }
+
+        if (! commit)
+          frozenLeqs.backtrack();
+
+        /*
+        if (tentative && !commit)
+          {
+            // These type variables used to be free. Since we don't commit, 
+            // we must set them free again!
+            try{
+              for (Iterator i = formerFree.iterator(); i.hasNext();)
+                {
+                  Element e = (Element) i.next();
+                  e.setKind(null);
+                  floating.add(e);
+                }
+            }
+            finally{
+              formerFree.endOfIteration();
+            }
+          }
+        */
+      }
+
     soft.backtrack();
-    for(Iterator i = floating.iterator();i.hasNext();)
-      ((Element)i.next()).setKind(null);
-    floating.endOfIteration();
-    frozenLeqs.backtrack();
+    formerFree.backtrack();
+
+    if (!tentative && existentialLevel > 0)
+      existentialLevel--;
   }
 
+  /** Marker used to know how deep we are inside existential mode, so that
+      we know when to exit from it.
+  */
+  private static int existentialLevel = 0;
+
   public static void startSimplify()
   {
     for(Iterator i = constraints.iterator();
@@ -276,7 +324,11 @@ public abstract class Engine
   public static void register(Element e)
   {
     if(dbg) Debug.println("Registering "+e);
-    
+
+    if (e.isExistential())
+      if (existentialLevel == 0)
+        existentialLevel = 1;
+
     if(e.getKind()!=null)
       e.getKind().register(e);
     else
@@ -310,8 +362,9 @@ public abstract class Engine
   public static void tag(Element e, int variance)
   {
     Kind kind = e.getKind();
-    if(kind==null)
-      throw new InternalError("null kind for "+e);
+    if(kind==null) 
+      return;
+      //throw new InternalError("null kind for "+e);
     Engine.Constraint k = getConstraint(kind);
     if(k==null)
       throw new InternalError("null constraint for "+e);
@@ -380,8 +433,8 @@ public abstract class Engine
   public static Element canonify(Element e)
   {
     Kind kind = e.getKind();
-    if(kind==null)
-      throw new InternalError("null kind for "+e);
+    if(kind==null) return e;
+    //throw new InternalError("null kind for "+e);
     Engine.Constraint k = getConstraint(kind);
     if(k==null)
       throw new InternalError("null constraint for "+e);
@@ -420,7 +473,7 @@ public abstract class Engine
     while(!s.empty())
       {
 	Element e = (Element)s.pop();
-	
+
 	if(e.getKind()!=null)
 	  if(e.getKind()==k)
 	    continue;
@@ -429,6 +482,9 @@ public abstract class Engine
 	      ("Bad Kinding for " + e + ":\nhad: " + e.getKind() +
 	       "\nnew: " + k);
 	
+        if (e.isExistential())
+          formerFree.add(e);
+
 	// assert e.getKind()==null
 	k.register(e);
 	e.setKind(k);
@@ -544,6 +600,40 @@ public abstract class Engine
     }
     soft.clear();
 
+    boolean more;
+    do {
+      more = false;
+      try {
+        for(Iterator i = frozenLeqs.iterator(); i.hasNext();)
+          {
+            Leq leq = (Leq) i.next();
+            Element e1 = leq.e1;
+            Element e2 = leq.e2;
+
+            // If at least one of the two is existential, then we must
+            // keep both frozen.
+            if ((e1.isExistential() && ! e2.isExistential())
+                ||
+                (e2.isExistential() && ! e1.isExistential()))
+              {
+                // Since we are marking an element as existential, we should
+                // do one more pass to make sure all related elements are
+                // marked
+                more = true;
+
+                if (e1.isExistential())
+                  ((mlsub.typing.MonotypeVar) e2).setExistential();
+                else if (e2.isExistential())
+                  ((mlsub.typing.MonotypeVar) e1).setExistential();
+              }
+          }
+      }
+      finally{
+        frozenLeqs.endOfIteration();
+      }
+    }
+    while (more);
+
     try {
       for(Iterator i = floating.iterator();
 	  i.hasNext();)
@@ -551,31 +641,41 @@ public abstract class Engine
 	  Element e = (Element) i.next();
 	  
 	  // useful for nullness head on monotype vars
-	  if (e.getKind() != null)
+          // we don't set existential in stone either, because they
+          // might be put into a kind later on.
+	  if (e.getKind() != null || e.isExistential())
             continue;
 
 	  if(dbg) Debug.println("Registering variable "+e);
 
 	  e.setKind(variablesConstraint);
 	  variablesConstraint.register(e);
+          i.remove();
 	}
     }
     finally{
       floating.endOfIteration();
     }
-    floating.clear();
     
     try {
       for(Iterator i = frozenLeqs.iterator(); i.hasNext();)
 	{
 	  Leq leq = (Leq) i.next();
+          Element e1 = leq.e1;
+          Element e2 = leq.e2;
+
+          // By the above code, if e2 is existential, e1 was marked as
+          // existential too, so we don't need to test e2.
+          if (e1.isExistential())
+            continue;
+
 	  variablesConstraint.leq(leq.e1, leq.e2, initialContext);
+          i.remove();
 	}
     }
     finally {
       frozenLeqs.endOfIteration();
     }
-    frozenLeqs.clear();
   }
   
   /** Maps a Kind to its lowlevel constraint */
@@ -619,6 +719,9 @@ public abstract class Engine
   /** The elements that have not yet been rigidified. */
   private static final BackableList soft = new BackableList();
   
+  /** The elements that have been put into a kind since the last mark. */
+  private static final BackableList formerFree = new BackableList();
+  
   /** The constraint of monotype variables */
   public static Engine.Constraint variablesConstraint;
   
@@ -691,7 +794,7 @@ public abstract class Engine
     private boolean variables = false;
     
     // this is not too logical to have this...
-    public mlsub.typing.Monotype freshMonotype()
+    public mlsub.typing.Monotype freshMonotype(boolean existential)
     {
       return null;
     }
@@ -894,9 +997,9 @@ public abstract class Engine
       k0.mark();
     }
 
-    void backtrack()
+    void backtrack(boolean ignore)
     {
-      k0.backtrack();
+      k0.backtrack(ignore);
     }
 
     void startSimplify()
